!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_dlsm_scalapack_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI), Yusuke HIROTA (YH)
!! @date    2016/02/29 (YH)
!
!  (c) Copyright 2014--2016 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_dlsm_scalapack_mod

  use kml_plugin_tools_mod

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

  ! callback procedures
  !
  public  :: KMATH_Plugin_Setup_DLSM_ScaLAPACK
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pctxt
    integer                     :: rank
    integer                     :: nprocs
    integer                     :: nproc_row
    integer                     :: nproc_col
    integer                     :: myproc_row
    integer                     :: myproc_col
    integer                     :: nsize
    integer                     :: msize
    integer                     :: nrow_loc
    integer                     :: ncol_loc_a
    integer                     :: ncol_loc_bx
    integer                     :: lda
    integer                     :: ldb
    integer                     :: ldx

    PetscScalar,        pointer :: A_loc(:) => null()
    PetscScalar,        pointer :: B_loc(:) => null()
    PetscScalar,        pointer :: X_loc(:) => null()

    integer                     :: desca(9)
    integer                     :: descb(9)
    integer                     :: descx(9)

    integer                     :: blacs_ctxt

    type(s_pctxt),      pointer :: self  => null()

  end type s_pctxt

  integer, parameter :: blksize = 64

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_DLSM_ScaLAPACK
  !> @brief      Plug-in setup routine
  !! @authors    NT
  !! @param[in]  context     context
  !! @date       2016/02/29 (YH)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_DLSM_ScaLAPACK(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pctxt),   pointer :: pctxt


#if defined(DEBUG)
    write(6,*) "KMATH_Plugin_Setup_DLSM_ScaLAPACK> called."
#endif

    allocate(pctxt)
    pctxt%self => pctxt

    call kml_regist_plugin_routines(context,            &
                                    pctxt,              &
                                    'A,B',              &
                                    'X',                &
                                    'A,B,X',            &
                                    kmp_initialize,     &
                                    kmp_shutdown,       &
                                    kmp_realloc,        &
                                    kmp_solve,          &
                                    kmp_to_internal,    &
                                    kmp_to_standard,    &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_DLSM_ScaLAPACK

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/19 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Initialize> called."
#endif

    call blacs_pinfo(pctxt%rank, pctxt%nprocs)

    call kml_get_row_col(pctxt%nprocs, pctxt%nproc_row, pctxt%nproc_col)

    call blacs_get(0, 0, pctxt%blacs_ctxt)

    call blacs_gridinit(pctxt%blacs_ctxt, 'R', &
                        pctxt%nproc_row, pctxt%nproc_col)

    call blacs_gridinfo(pctxt%blacs_ctxt, &
                        pctxt%nproc_row,  pctxt%nproc_col, &
                        pctxt%myproc_row, pctxt%myproc_col)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/19 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Shutdown> called."
#endif

    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%B_loc)
      deallocate(pctxt%X_loc)
    end if

    call blacs_gridexit(pctxt%blacs_ctxt)

    deallocate(pctxt%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/19 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    Mat                      :: B
    integer                  :: m, n, ierr, info

    ! function
    integer                  :: numroc


#if defined(DEBUG)
    write(6,*) "Kmp_Realloc> called. "
#endif

    ! deallocate previous data
    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%B_loc)
      deallocate(pctxt%X_loc)
    end if

    ! check matrix size
    call kml_get_parameter_petsc(context, 'B', B, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call MatGetSize(B, n, m, ierr)

    pctxt%nsize = n
    pctxt%msize = m
    pctxt%nrow_loc = numroc(pctxt%nsize, blksize, pctxt%myproc_row, 0, pctxt%nproc_row)
    pctxt%ncol_loc_a = numroc(pctxt%nsize, blksize, pctxt%myproc_col, 0, pctxt%nproc_col)
    pctxt%ncol_loc_bx = numroc(pctxt%msize, blksize, pctxt%myproc_col, 0, pctxt%nproc_col)

    if (pctxt%nrow_loc == 0) then
      pctxt%lda = blksize
    else
      pctxt%lda = ((pctxt%nrow_loc - 1) / blksize + 1) * blksize
    end if
    pctxt%ldb = pctxt%lda
    pctxt%ldx = pctxt%lda

    ! allocate internal memory
    allocate(pctxt%A_loc(max(1, pctxt%lda * pctxt%ncol_loc_a)))
    allocate(pctxt%B_loc(max(1, pctxt%ldb * pctxt%ncol_loc_bx)))
    allocate(pctxt%X_loc(max(1, pctxt%ldx * pctxt%ncol_loc_bx)))

    ! set descriptor for A
    call descinit(pctxt%desca, pctxt%nsize, pctxt%nsize, &
                  blksize, blksize, 0, 0, pctxt%blacs_ctxt, pctxt%lda, info)

    ! set descriptor for B
    call descinit(pctxt%descb, pctxt%nsize, pctxt%msize, &
                  blksize, blksize, 0, 0, pctxt%blacs_ctxt, pctxt%ldb, info)

    ! set descriptor for X
    call descinit(pctxt%descx, pctxt%nsize, pctxt%msize, &
                  blksize, blksize, 0, 0, pctxt%blacs_ctxt, pctxt%ldx, info)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/19 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: ierr, info
    character(20)            :: mat_type

    integer, allocatable     :: iwork1(:)


#if defined(DEBUG)
    write(6,*) "Kmp_Solve> called."
#endif

    mat_type = 'general'
    call kml_get_parameter_string(context, 'matrix_type', mat_type, ierr)
    if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalSolve)
      return 
    end if
    call kml_tolower_case(mat_type)

#if defined(PETSC_USE_COMPLEX)
    call pzlacpy('A', pctxt%nsize, pctxt%msize, pctxt%B_loc, 1, 1, pctxt%descb, &
                 pctxt%X_loc, 1, 1, pctxt%descx)
#else
    call pdlacpy('A', pctxt%nsize, pctxt%msize, pctxt%B_loc, 1, 1, pctxt%descb, &
                 pctxt%X_loc, 1, 1, pctxt%descx)
#endif
!    pctxt%X_loc(:) = pctxt%B_loc(:)

    if (mat_type == 'general') then

      allocate(iwork1(pctxt%desca(9) + blksize))

#if defined(PETSC_USE_COMPLEX)

      call pzgesv(pctxt%nsize, pctxt%msize, pctxt%A_loc, 1, 1, pctxt%desca, iwork1, &
                  pctxt%X_loc, 1, 1, pctxt%descx, info)

#else

      call pdgesv(pctxt%nsize, pctxt%msize, pctxt%A_loc, 1, 1, pctxt%desca, iwork1, &
                  pctxt%X_loc, 1, 1, pctxt%descx, info)

#endif
      deallocate(iwork1)

      if (info == 0) then
        call kml_set_error(error, EP_None)
      else
        call kml_set_error(error, EP_SolveFail)
      end if

    else ! symmetric

#if defined(PETSC_USE_COMPLEX)

      call pzposv('U', pctxt%nsize, pctxt%msize, pctxt%A_loc, 1, 1, pctxt%desca, &
                  pctxt%X_loc, 1, 1, pctxt%descx, info)

#else

      call pdposv('U', pctxt%nsize, pctxt%msize, pctxt%A_loc, 1, 1, pctxt%desca, &
                  pctxt%X_loc, 1, 1, pctxt%descx, info)

#endif

      if (info == 0) then
        call kml_set_error(error, EP_None)
      else
        call kml_set_error(error, EP_SolveFail)
      end if

    end if

    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  in_param    input parameter number which convert to internal
  !! @param[in]  in_petsc    actual PETSc input variable 
  !! @date       2015/02/19 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pctxt, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc

    ! local variables
    Mat                      :: A, Asub
    Mat                      :: B, Bsub
    IS                       :: isrow, iscol
    integer                  :: ierr, i, rowr, ncolsr

    PetscScalar, allocatable :: valuesr(:)
    integer,     allocatable :: idxrow(:)
    integer,     allocatable :: idxcol(:)
    integer,     allocatable :: colsr(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Internal> called.  IN:", in_param, " (", in_petsc, ")"
#endif

    if (in_param == 0) then

      ! parameter Mat:A

      call kml_get_parameter_petsc(context, 'A', A, ierr)
      if (ierr /= EI_None) then
        call kml_set_error(error, EP_FatalToInternal)
        return
      end if

      allocate(idxrow(max(1,pctxt%nrow_loc)))
      allocate(idxcol(max(1,pctxt%ncol_loc_a)))

      ! set local index
      do i = 0, pctxt%nsize - 1
        if (mod(i / blksize, pctxt%nproc_row) == pctxt%myproc_row) then
          idxrow((i / (blksize * pctxt%nproc_row)) * blksize + mod(i, blksize) + 1) = i
        end if
      end do
      do i = 0, pctxt%nsize - 1
        if (mod(i / blksize, pctxt%nproc_col) == pctxt%myproc_col) then
          idxcol((i / (blksize * pctxt%nproc_col)) * blksize + mod(i, blksize) + 1) = i
        end if
      end do

      ! set index set
      call ISCreateGeneral(PETSC_COMM_SELF, pctxt%nrow_loc, idxrow, &
                           PETSC_COPY_VALUES, isrow, ierr)
      call ISCreateGeneral(PETSC_COMM_SELF, pctxt%ncol_loc_a, idxcol, &
                           PETSC_COPY_VALUES, iscol, ierr)

      ! get submatrix
      call MatGetSubmatrices(A, 1, isrow, iscol, MAT_INITIAL_MATRIX, ASub,ierr)

      allocate(colsr  (max(1,pctxt%ncol_loc_a)))
      allocate(valuesr(max(1,pctxt%ncol_loc_a)))

      pctxt%A_loc = 0.0d0

      do rowr = 0, pctxt%nrow_loc - 1
        call MatGetRow(ASub, rowr, ncolsr, colsr, valuesr, ierr)
        do i = 1, ncolsr
          pctxt%A_loc(colsr(i)*pctxt%lda + rowr + 1) = valuesr(i)
        end do
        call MatRestoreRow(ASub, rowr, ncolsr, colsr, valuesr, ierr)
      end do

      ! destroy variables
      deallocate(colsr)
      deallocate(valuesr)
      deallocate(idxrow)
      deallocate(idxcol)

      call ISDestroy(isrow, ierr)
      call ISDestroy(iscol, ierr)

      call MatDestroy(ASub, ierr)

    else

      ! parameter Mat:B

      call kml_get_parameter_petsc(context, 'B', B, ierr)
      if (ierr /= EI_None) then
        call kml_set_error(error, EP_FatalToInternal)
        return
      end if

      allocate(idxrow(max(1,pctxt%nrow_loc)))
      allocate(idxcol(max(1,pctxt%ncol_loc_bx)))

      ! set local index
      do i = 0, pctxt%nsize - 1
        if (mod(i / blksize, pctxt%nproc_row) == pctxt%myproc_row) then
          idxrow((i / (blksize * pctxt%nproc_row)) * blksize + mod(i, blksize) + 1) = i
        end if
      end do
      do i = 0, pctxt%msize - 1
        if (mod(i / blksize, pctxt%nproc_col) == pctxt%myproc_col) then
          idxcol((i / (blksize * pctxt%nproc_col)) * blksize + mod(i, blksize) + 1) = i
        end if
      end do

      ! set index set
      call ISCreateGeneral(PETSC_COMM_SELF, pctxt%nrow_loc, idxrow, &
                           PETSC_COPY_VALUES, isrow, ierr)
      call ISCreateGeneral(PETSC_COMM_SELF, pctxt%ncol_loc_bx, idxcol, &
                           PETSC_COPY_VALUES, iscol, ierr)

      ! get submatrix
      call MatGetSubmatrices(B, 1, isrow, iscol, MAT_INITIAL_MATRIX, BSub,ierr)

      allocate(colsr  (max(1,pctxt%ncol_loc_bx)))
      allocate(valuesr(max(1,pctxt%ncol_loc_bx)))

      pctxt%B_loc = 0.0d0

      do rowr = 0, pctxt%nrow_loc - 1
        call MatGetRow(BSub, rowr, ncolsr, colsr, valuesr, ierr)
        do i = 1, ncolsr
          pctxt%B_loc(colsr(i)*pctxt%ldb + rowr + 1) = valuesr(i)
        end do
        call MatRestoreRow(BSub, rowr, ncolsr, colsr, valuesr, ierr)
      end do

      ! destroy variables
      deallocate(colsr)
      deallocate(valuesr)
      deallocate(idxrow)
      deallocate(idxcol)

      call ISDestroy(isrow, ierr)
      call ISDestroy(iscol, ierr)

      call MatDestroy(BSub, ierr)

    end if

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter which number convert to standard
  !! @param[in]  out_petsc   actual PETSc output variables
  !! @date       2015/02/19 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pctxt, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc

    ! local variables
    Mat                      :: X
    integer                  :: i, j, ierr, il, jl, iar(1), jar(1)
    PetscScalar              :: v(1)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Standard> called. OUT:", out_param, " (", out_petsc, ")"
#endif

    call kml_get_parameter_petsc(context, 'X', X, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalToStandard)
      return

    end if

    do i = 0, pctxt%nsize-1
      do j = 0, pctxt%msize-1
        if(mod(i / blksize, pctxt%nproc_row) == pctxt%myproc_row .and. &
           mod(j / blksize, pctxt%nproc_col) == pctxt%myproc_col) then
          iar(1) = i
          jar(1) = j
          il = (i / (blksize * pctxt%nproc_row)) * blksize + mod(i, blksize)
          jl = (j / (blksize * pctxt%nproc_col)) * blksize + mod(j, blksize)
          v(1) = pctxt%X_loc(il + jl * pctxt%ldx + 1)
          call MatSetValues(X, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
        end if
      end do
    end do

    call MatAssemblyBegin(X, MAT_FINAL_ASSEMBLY, ierr)
    call MatAssemblyEnd  (X, MAT_FINAL_ASSEMBLY, ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter as re-use source
  !! @param[in]  in_param    input parameter as re-use destination
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pctxt, error, out_param, in_param)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param


#if defined(DEBUG)
    write(6,*) "Kmp_Reuse_Internal> called. OUT:", out_param, &
               " => IN:", in_param
#endif

    if (out_param == 0 .and. in_param == 1) then

      ! out_param: 0  ->  'X'
      ! in_param:  1  ->  'B'

      pctxt%B_loc(:) = pctxt%X_loc(:)
      call kml_set_error(error, EP_None)

    else

      call kml_set_error(error, EP_FatalReuse)

    end if

    return

  end subroutine kmp_reuse_internal

end module kmath_plugin_dlsm_scalapack_mod
