!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_dsvd_scalapack_mod
!> @brief   kmath lib API module
!! @authors Yusuke Hirota (YH), Toshiyuki Imamura (TI)
!! @date    2016/01/29 (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_dsvd_scalapack_mod

  use kml_plugin_tools_mod

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

  ! callback procedures
  !
  public  :: KMATH_Plugin_Setup_DSVD_ScaLAPACK
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pctxt
    integer                     :: rank
    integer                     :: nprocs
    integer                     :: nproc_row
    integer                     :: nproc_col
    integer                     :: myproc_row
    integer                     :: myproc_col
    integer                     :: msize
    integer                     :: nsize
    integer                     :: minmn
    integer                     :: nrow_loc_a
    integer                     :: ncol_loc_a
    integer                     :: nrow_loc_u
    integer                     :: ncol_loc_u
    integer                     :: nrow_loc_vt
    integer                     :: ncol_loc_vt
    integer                     :: lda
    integer                     :: ldu
    integer                     :: ldvt
    integer                     :: mb = 64
    integer                     :: nb = 64

    PetscScalar,        pointer :: A_loc(:) => null()
    real(kind(0.0d0)),  pointer :: S_loc(:) => null()
    PetscScalar,        pointer :: U_loc(:) => null()
    PetscScalar,        pointer :: VT_loc(:) => null()

    integer                     :: desca(9)
    integer                     :: descu(9)
    integer                     :: descvt(9)

    integer                     :: blacs_ctxt

    type(s_pctxt),      pointer :: self  => null()
  end type s_pctxt

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_DSVD_ScaLAPACK
  !> @brief      Plug-in setup routine
  !! @authors    YH, TI
  !! @param[in]  context     context
  !! @date       2016/01/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_DSVD_ScaLAPACK(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pctxt),   pointer :: pctxt


#if defined(DEBUG)
    write(6,*) "KMATH_Plugin_Setup_DSVD_ScaLAPACK> called."
#endif

    allocate(pctxt)
    pctxt%self => pctxt

    call kml_regist_plugin_routines(context,            &
                                    pctxt,              &
                                    'A',                &
                                    'S,U,VT',           &
                                    'A,S,U,VT',         &
                                    kmp_initialize,     &
                                    kmp_shutdown,       &
                                    kmp_realloc,        &
                                    kmp_solve,          &
                                    kmp_to_internal,    &
                                    kmp_to_standard,    &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_DSVD_ScaLAPACK

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Initialize> called."
#endif

    call blacs_pinfo(pctxt%rank, pctxt%nprocs)

    call kml_get_row_col(pctxt%nprocs, pctxt%nproc_row, pctxt%nproc_col)

    call blacs_get(0, 0, pctxt%blacs_ctxt)

    call blacs_gridinit(pctxt%blacs_ctxt, 'R', &
                        pctxt%nproc_row, pctxt%nproc_col)

    call blacs_gridinfo(pctxt%blacs_ctxt, &
                        pctxt%nproc_row,  pctxt%nproc_col, &
                        pctxt%myproc_row, pctxt%myproc_col)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Shutdown> called."
#endif

    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%S_loc)
      deallocate(pctxt%U_loc)
      deallocate(pctxt%VT_loc)
    end if

    call blacs_gridexit(pctxt%blacs_ctxt)

    deallocate(pctxt%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    Mat                      :: A, U, VT
    integer                  :: mam, man, mum, mun, mvtm, mvtn, ierr, info

    ! function
    integer                  :: numroc


#if defined(DEBUG)
    write(6,*) "Kmp_Realloc> called. "
#endif

    ! deallocate previous data
    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%S_loc)
      deallocate(pctxt%U_loc)
      deallocate(pctxt%VT_loc)
    end if

    ! check matrix size
    call kml_get_parameter_petsc(context, 'A', A, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call kml_get_parameter_petsc(context, 'U', U, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call kml_get_parameter_petsc(context, 'VT', VT, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call MatGetSize(A, mam, man, ierr)
    pctxt%msize = mam
    pctxt%nsize = man
    pctxt%minmn = min(mam, man)

    call MatGetSize(U, mum, mun, ierr)
    if (mum /= pctxt%msize .or. mun /= pctxt%minmn) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    call MatGetSize(VT, mvtm, mvtn, ierr)
    if (mvtm /= pctxt%minmn .or. mvtn /= pctxt%nsize) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    ! determine the array dimension
    pctxt%nrow_loc_a = numroc(pctxt%msize, pctxt%mb, pctxt%myproc_row, 0, pctxt%nproc_row)
    pctxt%ncol_loc_a = numroc(pctxt%nsize, pctxt%nb, pctxt%myproc_col, 0, pctxt%nproc_col)
    if (pctxt%nrow_loc_a == 0) then
      pctxt%lda = pctxt%mb
    else
      pctxt%lda = ((pctxt%nrow_loc_a - 1) / pctxt%mb + 1) * pctxt%mb
    end if

    pctxt%nrow_loc_u = numroc(pctxt%msize, pctxt%mb, pctxt%myproc_row, 0, pctxt%nproc_row)
    pctxt%ncol_loc_u = numroc(pctxt%minmn, pctxt%nb, pctxt%myproc_col, 0, pctxt%nproc_col)
    if (pctxt%nrow_loc_u == 0) then
      pctxt%ldu = pctxt%mb
    else
      pctxt%ldu = ((pctxt%nrow_loc_u - 1) / pctxt%mb + 1) * pctxt%mb
    end if

    pctxt%nrow_loc_vt = numroc(pctxt%minmn, pctxt%mb, pctxt%myproc_row, 0, pctxt%nproc_row)
    pctxt%ncol_loc_vt = numroc(pctxt%nsize, pctxt%nb, pctxt%myproc_col, 0, pctxt%nproc_col)
    if (pctxt%nrow_loc_vt == 0) then
      pctxt%ldvt = pctxt%mb
    else
      pctxt%ldvt = ((pctxt%nrow_loc_vt - 1) / pctxt%mb + 1) * pctxt%mb
    end if

    ! allocate internal memory
    allocate(pctxt%A_loc(max(1, pctxt%lda * pctxt%ncol_loc_a)))
    allocate(pctxt%U_loc(max(1, pctxt%ldu * pctxt%ncol_loc_u)))
    allocate(pctxt%VT_loc(max(1, pctxt%ldvt * pctxt%ncol_loc_vt)))
    allocate(pctxt%S_loc(pctxt%minmn))

    ! set descriptor for A, U, VT
    call descinit(pctxt%desca, pctxt%msize, pctxt%nsize, &
                  pctxt%mb, pctxt%nb, 0, 0, pctxt%blacs_ctxt, pctxt%lda, info)
    call descinit(pctxt%descu, pctxt%msize, pctxt%minmn, &
                  pctxt%mb, pctxt%nb, 0, 0, pctxt%blacs_ctxt, pctxt%ldu, info)
    call descinit(pctxt%descvt, pctxt%minmn, pctxt%nsize, &
                  pctxt%mb, pctxt%nb, 0, 0, pctxt%blacs_ctxt, pctxt%ldvt, info)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                        :: lwork, info
    PetscScalar,       allocatable :: work1(:)
#if defined(PETSC_USE_COMPLEX)
    real(kind(0.0d0)), allocatable :: rwork1(:)
    integer                        :: lrwork
#endif


#if defined(DEBUG)
    write(6,*) "Kmp_Solve> called."
#endif

#if defined(PETSC_USE_COMPLEX)
    ! compute the SVD of a real matrix
    !

    ! first get work size
    lwork = -1
    allocate(work1(1))
    allocate(rwork1(1))
    call pzgesvd('V', 'V', pctxt%msize, pctxt%nsize, &
                 pctxt%A_loc, 1, 1, pctxt%desca, pctxt%S_loc, &
                 pctxt%U_loc, 1, 1, pctxt%descu, pctxt%VT_loc, 1, 1, pctxt%descvt, work1, lwork, rwork1, info)

    ! next actual calculation
    lwork = int(work1(1)) + 1
    lrwork = int(rwork1(1)) + 1
    deallocate(work1)
    deallocate(rwork1)
    allocate(rwork1(lrwork))
    allocate(work1(lwork))
    call pzgesvd('V', 'V', pctxt%msize, pctxt%nsize, &
                 pctxt%A_loc, 1, 1, pctxt%desca, pctxt%S_loc, &
                 pctxt%U_loc, 1, 1, pctxt%descu, pctxt%VT_loc, 1, 1, pctxt%descvt, work1, lwork, rwork1, info)

    deallocate(work1)
    deallocate(rwork1)

#else
    ! compute the SVD of a real matrix
    !

    ! first get work size
    lwork = -1
    allocate(work1(1))
    call pdgesvd('V', 'V', pctxt%msize, pctxt%nsize, &
                 pctxt%A_loc, 1, 1, pctxt%desca, pctxt%S_loc, &
                 pctxt%U_loc, 1, 1, pctxt%descu, pctxt%VT_loc, 1, 1, pctxt%descvt, work1, lwork, info)

    ! next actual calculation
    lwork = int(work1(1)) + 1
    deallocate(work1)
    allocate(work1(lwork))
    call pdgesvd('V', 'V', pctxt%msize, pctxt%nsize, &
                 pctxt%A_loc, 1, 1, pctxt%desca, pctxt%S_loc, &
                 pctxt%U_loc, 1, 1, pctxt%descu, pctxt%VT_loc, 1, 1, pctxt%descvt, work1, lwork, info)

    deallocate(work1)

#endif

    if (info == 0) then
      call kml_set_error(error, EP_None)
    else
      call kml_set_error(error, EP_SolveFail)
    end if

    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  in_param    input parameter number which convert to internal
  !! @param[in]  in_petsc    actual PETSc input variable 
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pctxt, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc

    ! local variables
    Mat                      :: A, Asub
    IS                       :: isrow, iscol
    integer                  :: ierr, i, rowr, ncolsr

    PetscScalar, allocatable :: valuesr(:)
    integer,     allocatable :: idxrow(:) ! Local index (i, j) is mapped to global index (idxrow(i), idxcol(j)).
    integer,     allocatable :: idxcol(:)
    integer,     allocatable :: colsr(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Internal> called.  IN:", in_param, " (", in_petsc, ")"
#endif

    call kml_get_parameter_petsc(context, 'A', A, ierr)
    if (ierr /= EI_None) then
      call kml_set_error(error, EP_FatalToInternal)
      return
    end if

    allocate(idxrow(max(1, pctxt%nrow_loc_a)))
    allocate(idxcol(max(1, pctxt%ncol_loc_a)))

    ! set local index
    do i = 0, pctxt%msize - 1
      if (mod(i / pctxt%mb, pctxt%nproc_row) == pctxt%myproc_row) then
        idxrow((i / (pctxt%nb * pctxt%nproc_row)) * pctxt%mb + mod(i, pctxt%mb) + 1) = i
      end if
    end do
    do i = 0, pctxt%nsize - 1
      if (mod(i / pctxt%nb, pctxt%nproc_col) == pctxt%myproc_col) then
        idxcol((i / (pctxt%nb * pctxt%nproc_col)) * pctxt%nb + mod(i, pctxt%nb) + 1) = i
      end if
    end do

    ! set index set
    call ISCreateGeneral(PETSC_COMM_SELF, pctxt%nrow_loc_a, idxrow, &
                         PETSC_COPY_VALUES, isrow, ierr)
    call ISCreateGeneral(PETSC_COMM_SELF, pctxt%ncol_loc_a, idxcol, &
                         PETSC_COPY_VALUES, iscol, ierr)

    ! get submatrix
    call MatGetSubmatrices(A, 1, isrow, iscol, MAT_INITIAL_MATRIX, ASub, ierr)

    allocate(colsr  (max(1,pctxt%ncol_loc_a)))
    allocate(valuesr(max(1,pctxt%ncol_loc_a)))

    pctxt%A_loc = 0.0d0

    do rowr = 0, pctxt%nrow_loc_a - 1
      call MatGetRow(ASub, rowr, ncolsr, colsr, valuesr, ierr)
      do i = 1, ncolsr
        pctxt%A_loc(colsr(i)*pctxt%lda + rowr + 1) = valuesr(i)
      end do
      call MatRestoreRow(ASub, rowr, ncolsr, colsr, valuesr, ierr)
    end do

    ! destroy variables
    deallocate(colsr)
    deallocate(valuesr)
    deallocate(idxrow)
    deallocate(idxcol)

    call ISDestroy(isrow, ierr)
    call ISDestroy(iscol, ierr)

    call MatDestroy(ASub, ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter which number convert to standard
  !! @param[in]  out_petsc   actual PETSc output variables
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pctxt, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc

    ! local variables
    Vec                      :: S
    Mat                      :: U, VT
    PetscScalar              :: v(1)
    PetscScalar, allocatable :: S_PetscScalar(:)
    integer                  :: i, j, il, jl, ierr, iar(1), jar(1)

    integer,     allocatable :: idxrow(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Standard> called. OUT:", out_param, " (", out_petsc, ")"
#endif

    if (out_param == 0) then

      ! parameter Vec:S

      call kml_get_parameter_petsc(context, 'S', S, ierr)
      if (ierr == EI_ParamNotFound) then
        call kml_set_error(error, EP_None)
        return

      else if (ierr == EI_FatalGetParam) then
        call kml_set_error(error, EP_FatalToStandard)
        return

      end if

      allocate(idxrow(pctxt%minmn))

      ! set local index
      do i = 0, pctxt%minmn - 1
        idxrow(i + 1) = i
      end do

      ! set value
      allocate(S_PetscScalar(pctxt%minmn))
#if defined(PETSC_USE_COMPLEX)
      S_PetscScalar(1 : pctxt%minmn) = cmplx(pctxt%S_loc(1 : pctxt%minmn), 0.0d0, kind(0.0d0))
#else
      S_PetscScalar(1 : pctxt%minmn) = pctxt%S_loc(1 : pctxt%minmn)
#endif
      if (pctxt%rank == 0) then
        call VecSetValues(S, pctxt%minmn, idxrow, S_PetscScalar, &
                          INSERT_VALUES,ierr)
      end if
      call VecAssemblyBegin(S, ierr)
      call VecAssemblyEnd  (S, ierr)

      deallocate(S_PetscScalar)
      deallocate(idxrow)

    else if (out_param == 1) then

      ! parameter Mat:U

      call kml_get_parameter_petsc(context, 'U', U, ierr)
      if (ierr == EI_ParamNotFound) then
        call kml_set_error(error, EP_None)
        return

      else if (ierr == EI_FatalGetParam) then
        call kml_set_error(error, EP_FatalToStandard)
        return

      end if

      do i = 0, pctxt%msize-1
        do j = 0, pctxt%minmn-1
          if(mod(i / pctxt%mb, pctxt%nproc_row) == pctxt%myproc_row .and. &
             mod(j / pctxt%nb, pctxt%nproc_col) == pctxt%myproc_col) then
            iar(1) = i
            jar(1) = j
            il = (i / (pctxt%mb * pctxt%nproc_row)) * pctxt%mb + mod(i, pctxt%mb)
            jl = (j / (pctxt%nb * pctxt%nproc_col)) * pctxt%nb + mod(j, pctxt%nb)
            v(1) = pctxt%U_loc(il + jl * pctxt%ldu + 1)
            call MatSetValues(U, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
          end if
        end do
      end do

      call MatAssemblyBegin(U, MAT_FINAL_ASSEMBLY, ierr)
      call MatAssemblyEnd  (U, MAT_FINAL_ASSEMBLY, ierr)

    else

      ! parameter Mat:VT

      call kml_get_parameter_petsc(context, 'VT', VT, ierr)
      if (ierr == EI_ParamNotFound) then
        call kml_set_error(error, EP_None)
        return

      else if (ierr == EI_FatalGetParam) then
        call kml_set_error(error, EP_FatalToStandard)
        return

      end if

      do i = 0, pctxt%minmn-1
        do j = 0, pctxt%nsize-1
          if(mod(i / 64, pctxt%nproc_row) == pctxt%myproc_row .and. &
             mod(j / 64, pctxt%nproc_col) == pctxt%myproc_col) then
            iar(1) = i
            jar(1) = j
            il = (i / (pctxt%mb * pctxt%nproc_row)) * pctxt%mb + mod(i, pctxt%mb)
            jl = (j / (pctxt%nb * pctxt%nproc_col)) * pctxt%nb + mod(j, pctxt%nb)
            v(1) = pctxt%VT_loc(il + jl * pctxt%ldvt + 1)
            call MatSetValues(VT, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
          end if
        end do
      end do

      call MatAssemblyBegin(VT, MAT_FINAL_ASSEMBLY, ierr)
      call MatAssemblyEnd  (VT, MAT_FINAL_ASSEMBLY, ierr)

    end if

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter as re-use source
  !! @param[in]  in_param    input parameter as re-use destination
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pctxt, error, out_param, in_param)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param

#if defined(DEBUG)
    write(6,*) "Kmp_Reuse_Internal> called. OUT:", out_param, &
               " => IN:", in_param
#endif

    if (out_param == 1 .and. in_param == 0) then

      ! out_param: 1  ->  'U'
      ! in_param:  0  ->  'A'

      pctxt%A_loc(:) = pctxt%U_loc(:)
      call kml_set_error(error, EP_None)

    else if (out_param == 2 .and. in_param == 0) then

      ! out_param: 2  ->  'VT'
      ! in_param:  0  ->  'A'

      pctxt%A_loc(:) = pctxt%VT_loc(:)
      call kml_set_error(error, EP_None)

    else

      call kml_set_error(error, EP_FatalReuse)

    end if

    return

  end subroutine kmp_reuse_internal

end module kmath_plugin_dsvd_scalapack_mod
