program test

  use kmath_lib_mod
  use kmath_plugin_deig_lapack_mod
  use kmath_plugin_deig_ssl2_mod

  implicit none

#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

  Mat :: A
  Mat :: Z1
  Mat :: Z2
  Vec :: X1
  Vec :: X2
  Vec :: EI
  Vec :: ZI
  Vec :: MXIZI
  Vec :: RESI
#if defined(PETSC_USE_COMPLEX)
  PetscScalar, parameter :: zero = cmplx(0.0d0, 0.0d0, kind(0.0d0))
  PetscScalar, parameter :: one = cmplx(1.0d0, 0.0d0, kind(0.0d0))
#else
  PetscScalar, parameter :: zero = 0.0d0
  PetscScalar, parameter :: one = 1.0d0
#endif

  PetscInt             :: ierr
  PetscScalar          :: v(1)
  PetscScalar          :: r1
  PetscReal            :: r
  PetscScalar, allocatable :: X1row(:), X2row(:)
  integer              :: sn
  integer              :: rank
  integer              :: m, n, i, j, iar(1), jar(1)
  integer, allocatable :: seed(:), idx(:)
  type(s_context)      :: h


  call random_seed(size = sn)
  allocate(seed(sn))
  seed = 123;
  call random_seed(put = seed)
  deallocate(seed)

  call PetscInitialize(PETSC_NULL_CHARACTER, ierr)

  call MPI_Comm_rank(PETSC_COMM_WORLD,rank,ierr)

  ! create matrix & vector
  m = 1000
  n = 1000

  call MatCreateDense(PETSC_COMM_WORLD, PETSC_DECIDE, &
       PETSC_DECIDE, m, n, PETSC_NULL_SCALAR,A, ierr)

  call VecCreate     (PETSC_COMM_WORLD, X1, ierr)
  call VecSetType    (X1, VECMPI, ierr)
  call VecSetSizes   (X1, PETSC_DECIDE, n, ierr)

  call MatCreateDense(PETSC_COMM_WORLD, PETSC_DECIDE, &
       PETSC_DECIDE, m, n, PETSC_NULL_SCALAR, Z1, ierr)

  call VecCreate     (PETSC_COMM_WORLD, X2, ierr)
  call VecSetType    (X2, VECMPI, ierr)
  call VecSetSizes   (X2, PETSC_DECIDE, n, ierr)

  call MatCreateDense(PETSC_COMM_WORLD, PETSC_DECIDE, &
       PETSC_DECIDE, m, n, PETSC_NULL_SCALAR, Z2, ierr)

  ! set symmetric matrix
  if (rank == 0) then
    do i = 0, m - 1
      iar(1) = i
      do j = 0, i
        jar(1) = j
        call random_number(r)
        r1 = r
        v(1) = r1
        call MatSetValues(A, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
        if(i /= j) &
          call MatSetValues(A, 1, jar, 1, iar, v, INSERT_VALUES, ierr)
      end do
    end do
  end if

  call MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY,ierr)
  call MatAssemblyEnd  (A,MAT_FINAL_ASSEMBLY,ierr)

  ! solve by KMATH API (with LAPACK)
  call KMATH_Create(h, PETSC_COMM_WORLD, ierr)

  call KMATH_Set_Plugin(h, KMATH_Plugin_Setup_DEIG_LAPACK, ierr)
  call KMATH_Set_Parameter(h, 'A', A,  ierr)
  call KMATH_Set_Parameter(h, 'X', X1, ierr)
  call KMATH_Set_Parameter(h, 'Z', Z1, ierr)
  call KMATH_Solve(h, ierr)

  ! solve by KMATH API (with SSL2)
  call KMATH_Set_Plugin(h, KMATH_Plugin_Setup_DEIG_SSL2, ierr)
  call KMATH_Set_Parameter(h, 'A', A,  ierr)
  call KMATH_Set_Parameter(h, 'X', X2, ierr)
  call KMATH_Set_Parameter(h, 'Z', Z2, ierr)
  call KMATH_Solve(h, ierr)

  call KMATH_Destroy(h, ierr)


  ! Show eigenvalues
  if (rank == 0) then
    print*,'*** The order of the eigenvalues in SSL2 is different from that in LAPACK. ***'
  end if
  do i = n, 1, - n / 5
    ! Create the elementary vector \bm{e}_1.
    call VecDuplicate(X1, EI, ierr)
    call VecSet(EI, zero, ierr)
    call VecSetValue(EI, i - 1, one, INSERT_VALUES, ierr)
    call VecAssemblyBegin(EI, ierr)
    call VecAssemblyEnd(EI, ierr)
    ! Get and show i-th eigenvalue (= Z EI).
    call VecDot(X1, EI, r, ierr)
    if (rank == 0) then
      print*, 'i = ', i - 1, ', eigenvalue (LAPACK): ', r
    end if
    call VecDot(X2, EI, r, ierr)
    if (rank == 0) then
      print*, 'i = ', i - 1, ', eigenvalue (SSL II): ', r
    end if
    ! Free variables.
    call VecDestroy(EI, ierr)
  enddo

  ! Residual check
  do i = 1, n, n / 3
    !!! LAPACK
    ! Create the elementary vector \bm{e}_1.
    call VecDuplicate(X1, EI, ierr)
    call VecSet(EI, zero, ierr)
    call VecSetValue(EI, i - 1, one, INSERT_VALUES, ierr)
    call VecAssemblyBegin(EI, ierr)
    call VecAssemblyEnd(EI, ierr)
    ! Get i-th eigenvalue (= Z EI).
    call VecDot(X1, EI, r, ierr)
    ! Get i-th eigenvector ZI (= Z EI).
    call VecDuplicate(X1, ZI, ierr)
    call MatMult(Z1, EI, ZI, ierr)
    ! Get i-th residual norm
    call VecDuplicate(ZI, MXIZI, ierr)
    call VecDuplicate(ZI, RESI, ierr)
    call VecCopy(ZI, MXIZI, ierr)
    call VecScale(MXIZI, -r, ierr)  ! - XI ZI
    call MatMultAdd(A, ZI, MXIZI, RESI, ierr) ! RESI = A ZI + (- XI ZI)
    ! Get and show i-th residual norm
    call VecNorm(RESI, NORM_2, r, ierr)
    if (rank == 0) then
      print*, 'i = ', i - 1, ', residual norm (LAPACK): ', r
    end if

    !!! SSL II
    ! Get i-th eigenvalue (= Z EI).
    call VecDot(X2, EI, r, ierr)
    ! Get i-th eigenvector ZI (= Z EI).
    call VecDuplicate(X2, ZI, ierr)
    call MatMult(Z2, EI, ZI, ierr)
    ! Get i-th residual norm
    call VecCopy(ZI, MXIZI, ierr)
    r = -r
    call VecScale(MXIZI, r, ierr)  ! - XI ZI
    call MatMultAdd(A, ZI, MXIZI, RESI, ierr) ! RESI = A ZI + (- XI ZI)
    ! Get and show i-th residual norm
    call VecNorm(RESI, NORM_2, r, ierr)
    if (rank == 0) then
      print*, 'i = ', i - 1, ', residual norm (SSL II): ', r
    end if

    ! Free variables.
    call VecDestroy(EI, ierr)
    call VecDestroy(ZI, ierr)
    call VecDestroy(MXIZI, ierr)
    call VecDestroy(EI, ierr)
  enddo

  call MatDestroy(A, ierr)
  call VecDestroy(X1,ierr)
  call MatDestroy(Z1,ierr)
  call VecDestroy(X2,ierr)
  call MatDestroy(Z2,ierr)

  call PetscFinalize(ierr)

end program test
